@@ -28,7 +28,7 @@ class GmailRepository(IMAPRepository):
     HOSTNAME = "imap.gmail.com"
     # Gmail IMAP server port
     PORT = 993
-    
+
     def __init__(self, reposname, account):
         """Initialize a GmailRepository object."""
         # Enforce SSL usage
@@ -36,6 +36,7 @@ class GmailRepository(IMAPRepository):
                                 'ssl', 'yes')
         IMAPRepository.__init__(self, reposname, account)
 
+
     def gethost(self):
         """Return the server name to connect to.
 
@@ -59,22 +60,15 @@ class GmailRepository(IMAPRepository):
 
     def getfolder(self, foldername):
         return self.getfoldertype()(self.imapserver, foldername,
-                                    self.nametrans(foldername),
-                                    self.accountname, self)
+                                    self)
 
     def getfoldertype(self):
         return folder.Gmail.GmailFolder
 
-    def getrealdelete(self, foldername):
-        # XXX: `foldername` is currently ignored - the `realdelete`
-        # setting is repository-wide
-        return self.getconfboolean('realdelete', 0)
-
     def gettrashfolder(self, foldername):
         #: Where deleted mail should be moved
         return  self.getconf('trashfolder','[Gmail]/Trash')
-	
+
     def getspamfolder(self):
         #: Gmail also deletes messages upon EXPUNGE in the Spam folder
         return  self.getconf('spamfolder','[Gmail]/Spam')
-
