@@ -1,5 +1,4 @@
-# Copyright (C) 2007 John Goerzen
-# <jgoerzen@complete.org>
+# Copyright (C) 2007-2011 John Goerzen & contributors
 #
 #    This program is free software; you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -14,54 +13,49 @@
 #    You should have received a copy of the GNU General Public License
 #    along with this program; if not, write to the Free Software
 #    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
-
-import urllib, sys, re, time, traceback, threading, thread
-from UIBase import UIBase
-from threading import *
+try:
+    from urllib import urlencode
+except ImportError: # python3
+    from urllib.parse import urlencode
+import sys
+import time
+import logging
+from threading import currentThread
+from offlineimap.ui.UIBase import UIBase
 import offlineimap
 
-protocol = '6.0.0'
+protocol = '7.0.0'
+
+class MachineLogFormatter(logging.Formatter):
+    """urlencodes any outputted line, to avoid multi-line output"""
+    def format(self, record):
+        # urlencode the "mesg" attribute and append to regular line...
+        line = super(MachineLogFormatter, self).format(record)
+        return line + urlencode([('', record.mesg)])[1:]
 
 class MachineUI(UIBase):
-    def __init__(s, config, verbose = 0):
-        UIBase.__init__(s, config, verbose)
-        s.safechars=" ;,./-_=+()[]"
-        s.iswaiting = 0
-        s.outputlock = Lock()
-        s._printData('__init__', protocol)
+    def __init__(self, config, loglevel = logging.INFO):
+        super(MachineUI, self).__init__(config, loglevel)
+        self._log_con_handler.createLock()
+        """lock needed to block on password input"""
+        # Set up the formatter that urlencodes the strings...
+        self._log_con_handler.setFormatter(MachineLogFormatter())
 
-    def isusable(s):
-        return True
+    def _printData(self, command, msg):
+        self.logger.info("%s:%s:%s" % (
+                'msg', command, currentThread().getName()), extra={'mesg': msg})
 
-    def _printData(s, command, data, dolock = True):
-        s._printDataOut('msg', command, data, dolock)
-
-    def _printWarn(s, command, data, dolock = True):
-        s._printDataOut('warn', command, data, dolock)
-
-    def _printDataOut(s, datatype, command, data, dolock = True):
-        if dolock:
-            s.outputlock.acquire()
-        try:
-            print "%s:%s:%s:%s" % \
-                    (datatype,
-                     urllib.quote(command, s.safechars), 
-                     urllib.quote(currentThread().getName(), s.safechars),
-                     urllib.quote(data, s.safechars))
-            sys.stdout.flush()
-        finally:
-            if dolock:
-                s.outputlock.release()
-
-    def _display(s, msg):
+    def _msg(s, msg):
         s._printData('_display', msg)
 
-    def warn(s, msg, minor = 0):
-        s._printData('warn', '%s\n%d' % (msg, int(minor)))
+    def warn(self, msg, minor = 0):
+        # TODO, remove and cleanup the unused minor stuff
+        self.logger.warning("%s:%s:%s:%s" % (
+                'warn', '', currentThread().getName(), msg))
 
-    def registerthread(s, account):
-        UIBase.registerthread(s, account)
-        s._printData('registerthread', account)
+    def registerthread(self, account):
+        super(MachineUI, self).registerthread(account)
+        self._printData('registerthread', account)
 
     def unregisterthread(s, thread):
         UIBase.unregisterthread(s, thread)
@@ -79,13 +73,13 @@ class MachineUI(UIBase):
     def validityproblem(s, folder):
         s._printData('validityproblem', "%s\n%s\n%s\n%s" % \
                 (folder.getname(), folder.getrepository().getname(),
-                 folder.getsaveduidvalidity(), folder.getuidvalidity()))
+                 folder.get_saveduidvalidity(), folder.get_uidvalidity()))
 
     def connecting(s, hostname, port):
         s._printData('connecting', "%s\n%s" % (hostname, str(port)))
 
     def syncfolders(s, srcrepos, destrepos):
-        s._printData('syncfolders', "%s\n%s" % (s.getnicename(srcrepos), 
+        s._printData('syncfolders', "%s\n%s" % (s.getnicename(srcrepos),
                                                 s.getnicename(destrepos)))
 
     def syncingfolder(s, srcrepos, srcfolder, destrepos, destfolder):
@@ -106,17 +100,14 @@ class MachineUI(UIBase):
                 (s.getnicename(sr), sf.getname(), s.getnicename(dr),
                  df.getname()))
 
-    def copyingmessage(s, uid, src, destlist):
-        ds = s.folderlist(destlist)
-        s._printData('copyingmessage', "%d\n%s\n%s\n%s"  % \
-                (uid, s.getnicename(src), src.getname(), ds))
-        
+    def copyingmessage(self, uid, num, num_to_copy, srcfolder, destfolder):
+        self._printData('copyingmessage', "%d\n%s\n%s\n%s[%s]" % \
+                (uid, self.getnicename(srcfolder), srcfolder.getname(),
+                 self.getnicename(destfolder), destfolder))
+
     def folderlist(s, list):
         return ("\f".join(["%s\t%s" % (s.getnicename(x), x.getname()) for x in list]))
 
-    def deletingmessage(s, uid, destlist):
-        s.deletingmessages(s, [uid], destlist)
-
     def uidlist(s, list):
         return ("\f".join([str(u) for u in list]))
 
@@ -124,24 +115,21 @@ class MachineUI(UIBase):
         ds = s.folderlist(destlist)
         s._printData('deletingmessages', "%s\n%s" % (s.uidlist(uidlist), ds))
 
-    def addingflags(s, uidlist, flags, destlist):
-        ds = s.folderlist(destlist)
+    def addingflags(s, uidlist, flags, dest):
         s._printData("addingflags", "%s\n%s\n%s" % (s.uidlist(uidlist),
                                                     "\f".join(flags),
-                                                    ds))
+                                                    dest))
 
-    def deletingflags(s, uidlist, flags, destlist):
-        ds = s.folderlist(destlist)
+    def deletingflags(s, uidlist, flags, dest):
         s._printData('deletingflags', "%s\n%s\n%s" % (s.uidlist(uidlist),
                                                       "\f".join(flags),
-                                                      ds))
+                                                      dest))
 
-    def threadException(s, thread):
-        print s.getThreadExceptionString(thread)
-        s._printData('threadException', "%s\n%s" % \
-                     (thread.getName(), s.getThreadExceptionString(thread)))
-        s.delThreadDebugLog(thread)
-        s.terminate(100)
+    def threadException(self, thread):
+        self._printData('threadException', "%s\n%s" % \
+                     (thread.getName(), self.getThreadExceptionString(thread)))
+        self.delThreadDebugLog(thread)
+        self.terminate(100)
 
     def terminate(s, exitstatus = 0, errortitle = '', errormsg = ''):
         s._printData('terminate', "%d\n%s\n%s" % (exitstatus, errortitle, errormsg))
@@ -161,19 +149,21 @@ class MachineUI(UIBase):
         return 0
 
 
-    def getpass(s, accountname, config, errmsg = None):
-        s.outputlock.acquire()
+    def getpass(self, accountname, config, errmsg = None):
+        if errmsg:
+            self._printData('getpasserror', "%s\n%s" % (accountname, errmsg),
+                         False)
+
+        self._log_con_handler.acquire() # lock the console output
         try:
-            if errmsg:
-                s._printData('getpasserror', "%s\n%s" % (accountname, errmsg),
-                             False)
-            s._printData('getpass', accountname, False)
+            self._printData('getpass', accountname, False)
             return (sys.stdin.readline()[:-1])
         finally:
-            s.outputlock.release()
+            self._log_con_handler.release()
 
-    def init_banner(s):
-        s._printData('initbanner', offlineimap.banner)
+    def init_banner(self):
+        self._printData('protocol', protocol)
+        self._printData('initbanner', offlineimap.banner)
 
-    def callhook(s, msg):
-        s._printData('callhook', msg)
+    def callhook(self, msg):
+        self._printData('callhook', msg)
